Index: m4/stl_hash.m4
===================================================================
--- m4/stl_hash.m4	(revision 386)
+++ m4/stl_hash.m4	(working copy)
@@ -14,28 +14,39 @@
   [AC_MSG_CHECKING(the location of hash_map)
   AC_LANG_SAVE
    AC_LANG_CPLUSPLUS
-   ac_cv_cxx_hash_map=""
-   for location in ext/hash_map hash_map; do
-     for namespace in __gnu_cxx "" std stdext; do
-       if test -z "$ac_cv_cxx_hash_map"; then
-         AC_TRY_COMPILE([#include <$location>],
-                        [${namespace}::hash_map<int, int> t],
-                        [ac_cv_cxx_hash_map="<$location>";
-                         ac_cv_cxx_hash_namespace="$namespace";])
-       fi
+   ac_cv_cxx_hash_map_header=""
+   ac_cv_cxx_hash_map_class=""
+   for location in [tr1/unordered_map ext/hash_map hash_map]; do
+     for namespace in [std::tr1 __gnu_cxx "" std stdext]; do
+       for name in [unordered_map hash_map]; do
+
+         if test -z "$ac_cv_cxx_hash_map_header"; then
+
+           AC_TRY_COMPILE([#include <$location>],
+                          [${namespace}::$name<int, int> t],
+                          [ac_cv_cxx_hash_map_header="<$location>";
+                          ac_cv_cxx_hash_namespace="$namespace";
+		       	  ac_cv_cxx_hash_map_class="$name";])
+         fi
+       done
      done
    done
-   ac_cv_cxx_hash_set=`echo "$ac_cv_cxx_hash_map" | sed s/map/set/`;
-   if test -n "$ac_cv_cxx_hash_map"; then
+   ac_cv_cxx_hash_set_header=`echo "$ac_cv_cxx_hash_map_header" | sed s/map/set/`;
+   ac_cv_cxx_hash_set_class=`echo "$ac_cv_cxx_hash_map_class" | sed s/map/set/`;
+   if test -n "$ac_cv_cxx_hash_map_header"; then
       AC_DEFINE(HAVE_HASH_MAP, 1, [define if the compiler has hash_map])
       AC_DEFINE(HAVE_HASH_SET, 1, [define if the compiler has hash_set])
-      AC_DEFINE_UNQUOTED(HASH_MAP_H,$ac_cv_cxx_hash_map,
+      AC_DEFINE_UNQUOTED(HASH_MAP_H,$ac_cv_cxx_hash_map_header,
                          [the location of <hash_map>])
-      AC_DEFINE_UNQUOTED(HASH_SET_H,$ac_cv_cxx_hash_set,
+      AC_DEFINE_UNQUOTED(HASH_SET_H,$ac_cv_cxx_hash_set_header,
                          [the location of <hash_set>])
+      AC_DEFINE_UNQUOTED(HASH_MAP_CLASS,$ac_cv_cxx_hash_map_class,
+                         [the name of <hash_set>])
+      AC_DEFINE_UNQUOTED(HASH_SET_CLASS,$ac_cv_cxx_hash_set_class,
+                         [the name of <hash_set>])
       AC_DEFINE_UNQUOTED(HASH_NAMESPACE,$ac_cv_cxx_hash_namespace,
                          [the namespace of hash_map/hash_set])
-      AC_MSG_RESULT([$ac_cv_cxx_hash_map])
+      AC_MSG_RESULT([$ac_cv_cxx_hash_map_header])
    else
       AC_MSG_RESULT()
       AC_MSG_WARN([could not find an STL hash_map])
Index: config.h.in
===================================================================
--- config.h.in	(revision 386)
+++ config.h.in	(working copy)
@@ -1,11 +1,17 @@
 /* config.h.in.  Generated from configure.ac by autoheader.  */
 
+/* the name of <hash_set> */
+#undef HASH_MAP_CLASS
+
 /* the location of <hash_map> */
 #undef HASH_MAP_H
 
 /* the namespace of hash_map/hash_set */
 #undef HASH_NAMESPACE
 
+/* the name of <hash_set> */
+#undef HASH_SET_CLASS
+
 /* the location of <hash_set> */
 #undef HASH_SET_H
 
Index: src/google/protobuf/stubs/hash.h
===================================================================
--- src/google/protobuf/stubs/hash.h	(revision 386)
+++ src/google/protobuf/stubs/hash.h	(working copy)
@@ -38,6 +38,7 @@
 #include <string.h>
 #include <google/protobuf/stubs/common.h>
 #include "config.h"
+#include <cstdio>
 
 #if defined(HAVE_HASH_MAP) && defined(HAVE_HASH_SET)
 #include HASH_MAP_H
@@ -134,67 +135,95 @@
 
 #else
 
+// Pull the hash from the discovered hash_map
 template <typename Key>
-struct hash : public HASH_NAMESPACE::hash<Key> {
+struct hash : public HASH_NAMESPACE::hash<Key> 
+{
 };
 
-template <typename Key>
-struct hash<const Key*> {
-  inline size_t operator()(const Key* key) const {
-    return reinterpret_cast<size_t>(key);
-  }
+// Implement a local hash. gcc 4.3+ have this very specialization, yet
+// it is not available in older releases
+template <>
+struct hash<string> 
+{
+    inline size_t operator()(const string &key) const 
+    {
+        const char *first = key.c_str();
+        size_t result = 0;
+        
+        for (size_t length = key.size(); length > 0; --length)
+            result = (result * 131) + *first++;
+                
+        return result;
+    }
+
+    static const size_t bucket_size = 4;
+    static const size_t min_buckets = 8;
+    
+    inline size_t operator()(const string &a, const string &b) const 
+    {
+        return a < b;
+    }
 };
 
+// STL that comes with gcc 3.4.x has a specialization for "const char *"
+// yet there is no common function name.
 template <>
-struct hash<const char*> : public HASH_NAMESPACE::hash<const char*> {
+struct hash<const char*> : public HASH_NAMESPACE::hash<const char *>
+{
+    inline size_t operator()(const char *key) const
+    {
+        return hash<std::string>()(std::string(key));
+    }
 };
 
+template <typename Key>
+struct hash<const Key *> 
+{
+    inline size_t operator()(const Key *key) const 
+    {
+        return reinterpret_cast<size_t>(key);
+    }
+};
+
 template <typename Key, typename Data,
           typename HashFcn = hash<Key>,
           typename EqualKey = std::equal_to<Key> >
-class hash_map : public HASH_NAMESPACE::hash_map<
-    Key, Data, HashFcn, EqualKey> {
+class hash_map 
+    : public HASH_NAMESPACE::HASH_MAP_CLASS<Key, Data, HashFcn, EqualKey>
+{
 };
 
 template <typename Key,
           typename HashFcn = hash<Key>,
           typename EqualKey = std::equal_to<Key> >
-class hash_set : public HASH_NAMESPACE::hash_set<
-    Key, HashFcn, EqualKey> {
+class hash_set 
+    : public HASH_NAMESPACE::HASH_SET_CLASS<Key, HashFcn, EqualKey> 
+{
 };
 
 #endif
 
-template <>
-struct hash<string> {
-  inline size_t operator()(const string& key) const {
-    return hash<const char*>()(key.c_str());
-  }
-
-  static const size_t bucket_size = 4;
-  static const size_t min_buckets = 8;
-  inline size_t operator()(const string& a, const string& b) const {
-    return a < b;
-  }
-};
-
 template <typename First, typename Second>
-struct hash<pair<First, Second> > {
-  inline size_t operator()(const pair<First, Second>& key) const {
-    size_t first_hash = hash<First>()(key.first);
-    size_t second_hash = hash<Second>()(key.second);
+struct hash<pair<First, Second> >
+{
+    inline size_t operator()(const pair<First, Second>& key) const
+    {
+        size_t first_hash = hash<First>()(key.first);
+        size_t second_hash = hash<Second>()(key.second);
+            
+        // FIXME(kenton):  What is the best way to compute this hash?  I have
+        // no idea!  This seems a bit better than an XOR.
+        return first_hash * ((1 << 16) - 1) + second_hash;
+    }
 
-    // FIXME(kenton):  What is the best way to compute this hash?  I have
-    // no idea!  This seems a bit better than an XOR.
-    return first_hash * ((1 << 16) - 1) + second_hash;
-  }
-
-  static const size_t bucket_size = 4;
-  static const size_t min_buckets = 8;
-  inline size_t operator()(const pair<First, Second>& a,
-                           const pair<First, Second>& b) const {
-    return a < b;
-  }
+    static const size_t bucket_size = 4;
+    static const size_t min_buckets = 8;
+    inline size_t operator()(const pair<First, Second>& a,
+                             const pair<First, Second>& b) const 
+    {
+        return a < b;
+    }
 };
 
 // Used by GCC/SGI STL only.  (Why isn't this provided by the standard
